@@ -1,6 +1,6 @@
 // Process server for soft ioc
 // David H. Thompson 8/29/2003
-// Ralph Lange 03/23/2010
+// Ralph Lange 03/05/2012
 // GNU Public License (GPLv3) applies - see www.gnu.org
 
 
@@ -29,6 +29,7 @@ const bool enableAllow = false;  // Default: NO
 #endif
 
 bool   inDebugMode;              // This enables a lot of printfs
+bool   inFgMode = false;         // This keeps child in the foreground, tty connected
 bool   logPortLocal;             // This restricts log port access to localhost
 bool   ctlPortLocal = true;      // Restrict control connections to localhost
 bool   autoRestart = true;       // Enables instant restart of exiting child
@@ -38,12 +39,15 @@ bool   quiet = false;            // Suppress info output (server)
 bool   setCoreSize = false;      // Set core size for child
 char   *procservName;            // The name of this beast (server)
 char   *childName;               // The name of that beast (child)
+char   *childExec;               // Exec to run as child
+char   **childArgv;              // Argv for child process
 int    connectionNo;             // Total number of connections
 char   *ignChars = NULL;         // Characters to ignore
 char   killChar = 0x18;          // Kill command character (default: ^X)
 char   toggleRestartChar = 0x14; // Toggle autorestart character (default: ^T)
 char   restartChar = 0x12;       // Restart character (default: ^R)
 char   quitChar = 0x11;          // Quit character (default: ^Q)
+char   logoutChar = 0x00;        // Logout client connection character (default: none)
 int    killSig = SIGKILL;        // Kill signal (default: SIGKILL)
 rlim_t coreSize;                 // Max core size for child
 char   *chDir;                   // Directory to change to before starting child
@@ -55,13 +59,16 @@ char   *pidFile;                 // File name for server PID
 char   defaultpidFile[] = "pid.txt";  // default
 char   *timeFormat;              // Time format string
 char   defaulttimeFormat[] = "%c";    // default
+bool   stampLog = false;         // Prefix log lines with time stamp
+char   *stampFormat;             // Log time stamp format string
 
 char   infoMessage1[512];        // Sign on message: server PID, child pwd and command line
 char   infoMessage2[128];        // Sign on message: child PID
 char   infoMessage3[128];        // Sign on message: available server commands
 
 char   *logFile = NULL;          // File name for log
-int    logFileFD=-1;;            // FD for log file
+int    logFileFD=-1;             // FD for log file
+int    logPort;                  // Port no. for logger connections
 int    debugFD=-1;               // FD for debug output
 
 #define MAX_CONNECTIONS 64
@@ -74,10 +81,13 @@ void OnPollTimeout();
 void forkAndGo();
 // Checks the command file (existence and access rights)
 int checkCommandFile(const char *command);
+void openLogFile();
+void ttySetCharNoEcho(bool save);
 
 // Signal handlers
 void OnSigPipe(int);
 void OnSigTerm(int);
+void OnSigHup(int);
 // Counter used for communication between sig handler and main()
 int sigPipeSet;
 
@@ -114,7 +124,7 @@ char getOptionChar ( const char* buf )
 
 void printUsage()
 {
-    printf("Usage: %s [options] <port> <command args ... >    (-h for help)\n",
+    printf("Usage: %s [options] <port> <command args ...>    (-h for help)\n",
            procservName);
 }
 
@@ -124,26 +134,30 @@ void printHelp()
     printf("<port>              use telnet <port> for command connections\n"
            "<command args ...>  command line to start child process\n"
            "Options:\n"
-           "    --allow           allow control connections from anywhere\n"
-           "    --autorestartcmd  command to toggle auto restart flag (^ for ctrl)\n"
-           "    --coresize <n>    sets maximum core size for child to <n>\n"
-           " -c --chdir <dir>     change directory to <dir> before starting child\n"
-           " -d --debug           enable debug mode (keeps child in foreground)\n"
-           " -h --help            print this message\n"
-           "    --holdoff <n>     set holdoff time between child restarts\n"
-           " -i --ignore <str>    ignore all chars in <str> (^ for ctrl)\n"
-           " -k --killcmd <str>   command to kill (reboot) the child (^ for ctrl)\n"
-           "    --killsig <n>     signal to send to child when killing\n"
-           " -l --logport <n>     allow log connections through telnet port <n>\n"
-           " -L --logfile <file>  write log to <file>\n"
-           " -n --name <str>      set child's name (defaults to command line)\n"
-           "    --noautorestart   do not restart child on exit by default\n"
-           " -p --pidfile <str>   name of PID file (for server PID)\n"
-           " -q --quiet           suppress informational output (server)\n"
-           "    --restrict        restrict log connections to localhost\n"
-           "    --timefmt <str>   set time format (strftime) to <str>\n"
-           " -V --version         print program version\n"
-           " -w --wait            wait for telnet cmd to manually start child\n"
+           "    --allow             allow control connections from anywhere\n"
+           "    --autorestartcmd    command to toggle auto restart flag (^ for ctrl)\n"
+           "    --coresize <n>      set maximum core size for child to <n>\n"
+           " -c --chdir <dir>       change directory to <dir> before starting child\n"
+           " -d --debug             debug mode (keeps child in foreground)\n"
+           " -e --exec <str>        specify child executable (default: arg0 of <command>)\n"
+           " -f --foreground        keep child in foreground (interactive)\n"
+           " -h --help              print this message\n"
+           "    --holdoff <n>       set holdoff time between child restarts\n"
+           " -i --ignore <str>      ignore all chars in <str> (^ for ctrl)\n"
+           " -k --killcmd <str>     command to kill (reboot) the child (^ for ctrl)\n"
+           "    --killsig <n>       signal to send to child when killing\n"
+           " -l --logport <n>       allow log connections through telnet port <n>\n"
+           " -L --logfile <file>    write log to <file>\n"
+           "    --logstamp [<str>]  prefix log lines with timestamp [strftime format]\n"
+           " -n --name <str>        set child's name (default: arg0 of <command>)\n"
+           "    --noautorestart     do not restart child on exit by default\n"
+           " -p --pidfile <str>     name of PID file (for server PID)\n"
+           " -q --quiet             suppress informational output (server)\n"
+           "    --restrict          restrict log connections to localhost\n"
+           "    --timefmt <str>     set time format (strftime) to <str>\n"
+           " -V --version           print program version\n"
+           " -w --wait              wait for telnet cmd to manually start child\n"
+           " -x --logoutcmd <str>   command to logout client connection (^ for ctrl)\n"
         );
 }
 
@@ -157,7 +171,7 @@ int main(int argc,char * argv[])
     int c;
     unsigned int i, j;
     int k;
-    int ctlPort, logPort = 0;
+    int ctlPort = 0;
     char *command;
     bool wrongOption = false;
     char buff[512];
@@ -172,7 +186,7 @@ int main(int argc,char * argv[])
     if ( !pidFile || strlen(pidFile) == 0 ) pidFile = defaultpidFile;
     if ( getenv("PROCSERV_DEBUG") != NULL ) inDebugMode = true;
 
-    const int ONE_CHAR_COMMANDS = 2;  // togglerestartcmd, killcmd
+    const int ONE_CHAR_COMMANDS = 3;  // togglerestartcmd, killcmd, logoutcmd
 
     while (1) {
         static struct option long_options[] = {
@@ -181,6 +195,8 @@ int main(int argc,char * argv[])
             {"coresize",       required_argument, 0, 'C'},
             {"chdir",          required_argument, 0, 'c'},
             {"debug",          no_argument,       0, 'd'},
+            {"exec",           required_argument, 0, 'e'},
+            {"foreground",     no_argument,       0, 'f'},
             {"help",           no_argument,       0, 'h'},
             {"holdoff",        required_argument, 0, 'H'},
             {"ignore",         required_argument, 0, 'i'},
@@ -188,6 +204,7 @@ int main(int argc,char * argv[])
             {"killsig",        required_argument, 0, 'K'},
             {"logport",        required_argument, 0, 'l'},
             {"logfile",        required_argument, 0, 'L'},
+            {"logstamp",       optional_argument, 0, 'S'},
             {"name",           required_argument, 0, 'n'},
             {"noautorestart",  no_argument,       0, 'N'},
             {"pidfile",        required_argument, 0, 'p'},
@@ -196,13 +213,14 @@ int main(int argc,char * argv[])
             {"timefmt",        required_argument, 0, 'F'},
             {"version",        no_argument,       0, 'V'},
             {"wait",           no_argument,       0, 'w'},
+            {"logoutcmd",      required_argument, 0, 'x'},
             {0, 0, 0, 0}
         };
 
         /* getopt_long stores the option index here. */
         int option_index = 0;
 
-        c = getopt_long (argc, argv, "+c:dhi:k:l:L:n:p:qVw",
+        c = getopt_long (argc, argv, "+c:de:fhi:k:l:L:n:p:qVwx:",
                          long_options, &option_index);
 
         /* Detect the end of the options. */
@@ -233,8 +251,22 @@ int main(int argc,char * argv[])
             inDebugMode = true;
             break;
 
+        case 'e':                                 // Child executable
+            childExec = strdup( optarg );
+            break;
+
+        case 'f':                                 // Foreground mode
+            inFgMode = true;
+            break;
+
         case 'F':                                 // Time string format
-            timeFormat = strdup( optarg );
+            timeFormat = strdup(optarg);
+            break;
+
+        case 'S':                                 // Log time stamp format
+            stampLog = true;
+            if (optarg)
+                stampFormat = strdup(optarg);
             break;
 
         case 'h':                                 // Help
@@ -319,6 +351,10 @@ int main(int argc,char * argv[])
             waitForManualStart = true;
             break;
 
+        case 'x':                                 // Logout command
+            logoutChar = getOptionChar(optarg);
+            break;
+
         case 'T':                                 // Toggle auto restart command
             toggleRestartChar = getOptionChar ( optarg );
             break;
@@ -345,34 +381,50 @@ int main(int argc,char * argv[])
     }
 
     // Single command characters should be ignored, too
-    if (ignChars == NULL && (killChar || toggleRestartChar))
+    if (ignChars == NULL && (killChar || toggleRestartChar || logoutChar))
         ignChars = (char*) calloc(1 + ONE_CHAR_COMMANDS, 1);
     if (killChar)
         strncat (ignChars, &killChar, 1);
     if (toggleRestartChar)
         strncat (ignChars, &toggleRestartChar, 1);
+    if (logoutChar)
+        strncat (ignChars, &logoutChar, 1);
 
     // Set up available server commands message
     PRINTF("Setting up messages\n");
-    sprintf( infoMessage3, "@@@ Use %s%c to restart the child now, %s%c to quit the server" NL,
-             restartChar < 32 ? "^" : "",
-             restartChar < 32 ? restartChar + 64 : restartChar,
-             quitChar < 32 ? "^" : "",
-             quitChar < 32 ? quitChar + 64 : quitChar);
+    sprintf(infoMessage3, "@@@ %s%c or %s%c restarts the child, %s%c quits the server",
+            CTL_SC(restartChar), CTL_SC(killChar), CTL_SC(quitChar));
+    if (logoutChar) {
+        sprintf(buff, ", %s%c closes this connection",
+                CTL_SC(logoutChar));
+        strcat(infoMessage3, buff);
+    }
+    strcat(infoMessage3, NL);
 
     ctlPort = atoi(argv[optind]);
     command = argv[optind+1];
-    if ( childName == NULL ) childName = command;
+
+    if (childName == NULL) childName = command;
+    childArgv = argv + optind;
+    if (childExec == NULL) {
+        childArgv++;
+        childExec = command;
+    }
 
     if ( ctlPort < 1024 )
     {
         fprintf( stderr,
                  "%s: invalid control port %d (<1024)\n",
                  procservName, ctlPort );
-	exit(1);
+        exit(1);
     }
 
-    if ( checkCommandFile( command ) ) exit( errno );
+    if (stampLog && !stampFormat) {
+        stampFormat = (char*) calloc(strlen(timeFormat)+4, 1);
+        sprintf(stampFormat, "[%s] ", timeFormat);
+    }
+
+    if (checkCommandFile(childExec)) exit(errno);
 
     struct sigaction sig;
     memset(&sig, 0, sizeof(sig));
@@ -382,6 +434,16 @@ int main(int argc,char * argv[])
     sigaction(SIGPIPE, &sig, NULL);
     sig.sa_handler = &OnSigTerm;
     sigaction(SIGTERM, &sig, NULL);
+    sig.sa_handler = &OnSigHup;
+    sigaction(SIGHUP, &sig, NULL);
+    sig.sa_handler = SIG_IGN;
+    sigaction(SIGXFSZ, &sig, NULL);
+    if (inFgMode) {
+        sig.sa_handler = SIG_IGN;
+        sigaction(SIGINT, &sig, NULL);
+        sig.sa_handler = SIG_IGN;
+        sigaction(SIGQUIT, &sig, NULL);
+    }
 
     // Make an accept item to listen for control connections
     PRINTF("Creating control listener\n");
@@ -417,26 +479,21 @@ int main(int argc,char * argv[])
 
     procservPid=getpid();
 
-    if ( inDebugMode == false )
+    openLogFile();
+
+    if (false == inFgMode && false == inDebugMode)
     {
-	forkAndGo();
-	writePidFile();
+        forkAndGo();
+        writePidFile();
     }
     else
     {
         debugFD = 1;          // Enable debug messages
     }
 
-    // Open log file
-    if ( logFile ) {
-        logFileFD = open( logFile, O_CREAT|O_WRONLY|O_APPEND, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH );
-        if ( logFileFD == -1 ) {         // Don't stop here - just go without
-            fprintf( stderr,
-                     "%s: unable to open log file %s\n",
-                     procservName, logFile );
-        } else {
-            PRINTF( "Opened file %s for logging\n", logFile );
-        }
+    if (inFgMode) {
+        ttySetCharNoEcho(true);
+        AddConnection(clientFactory(0));
     }
 
     // Record some useful data for managers 
@@ -489,9 +546,7 @@ int main(int argc,char * argv[])
 
         ready = select(nFd, &fdset, NULL, NULL, &timeout);
 
-        if (ready == -1) {                    // Error
-            perror("Error in select() call");
-        } else if (ready == 0) {              // Timeout
+        if (0 == ready) {                     // Timeout
             // Go clean up dead connections
             OnPollTimeout();
             connectionItem * npi; 
@@ -500,9 +555,13 @@ int main(int argc,char * argv[])
             // This call returns NULL if the process item lives
             if (processFactoryNeedsRestart())
             {
-                npi= processFactory(argc-optind-2,argv+optind+1);
+                npi= processFactory(childExec, childArgv);
                 if (npi) AddConnection(npi);
             }
+        } else if (-1 == ready) {             // Error
+            if (EINTR != errno) {
+                perror("Error in select() call");
+            }
         } else {                              // Work to be done
             // Loop through all connections
             p = connectionItem::head;
@@ -513,6 +572,7 @@ int main(int argc,char * argv[])
             OnPollTimeout();
         }
     }
+    ttySetCharNoEcho(false);
 }
 
 // Connection items call this to send messages to others
@@ -523,24 +583,40 @@ int main(int argc,char * argv[])
 void SendToAll(const char * message,int count,const connectionItem * sender)
 {
     connectionItem * p = connectionItem::head;
+    char stamp[64];
+    int len = 0;
+    time_t now;
+    struct tm now_tm;
 
+    if (stampLog) {
+        time(&now);
+        localtime_r(&now, &now_tm);
+        strftime(stamp, sizeof(stamp)-1, stampFormat, &now_tm);
+        len = strlen(stamp);
+    }
     // Log the traffic to file / stdout (debug)
-    if ( sender==NULL  || sender->isProcess() )
+    if (sender==NULL || sender->isProcess())
     {
-        if ( logFileFD > 0 ) write( logFileFD, message, count );
-        if ( debugFD > 0 ) write( debugFD, message, count );
+        if (logFileFD > 0) {
+            if (stampLog) write(logFileFD, stamp, len);
+            write(logFileFD, message, count);
+        }
+        if (false == inFgMode && debugFD > 0) write(debugFD, message, count);
     }
 
     while ( p ) {
 	if ( p->isProcess() )
 	{
 	    // Non-null senders that are not processes can send to processes
-	    if (sender && !sender->isProcess()) p->Send(message,count);
+        if (sender && !sender->isProcess()) p->Send(message, count);
 	}
 	else // Not a process
 	{
 	    // Null senders and processes can send to non-processes (ie connections)
-	    if (sender==NULL || sender->isProcess() ) p->Send(message,count);
+        if (sender==NULL || sender->isProcess()) {
+            if (stampLog && p->isLogger()) p->Send(stamp, len);
+            p->Send(message, count);
+        }
 	}
 	p=p->next;
     }
@@ -553,33 +629,33 @@ void OnPollTimeout()
     pid_t pid;
     int wstatus;
     connectionItem *pc, *pn;
-    char buf[128];
+    char buf[128] = NL;
 
     pid = waitpid(-1, &wstatus, WNOHANG);
     if (pid > 0 ) {
         pc = connectionItem::head;
-        while(pc)
-	{
+        while (pc) {
             pc->markDeadIfChildIs(pid);
             pc=pc->next;
-	}
+        }
 
-	sprintf( buf, NL "@@@ @@@ @@@ @@@ @@@" NL
-                 "@@@ Received a sigChild for process %ld." , (long) pid );
+        SendToAll(buf, strlen(buf), NULL);
+        strcpy(buf, "@@@ @@@ @@@ @@@ @@@" NL);
+        SendToAll(buf, strlen(buf), NULL);
 
-	if (WIFEXITED(wstatus))
-	{
-	    sprintf( buf + strlen(buf), " Normal exit status = %d",
-                     WEXITSTATUS(wstatus) );
-	}
+        sprintf(buf, "@@@ Received a sigChild for process %ld.", (long) pid);
 
-	if (WIFSIGNALED(wstatus))
-	{
-	    sprintf( buf + strlen(buf), " The process was killed by signal %d",
-                     WTERMSIG(wstatus) );
-	}
-	strcat( buf, NL );
-	SendToAll( buf, strlen(buf), NULL );
+        if (WIFEXITED(wstatus)) {
+            sprintf(buf + strlen(buf), " Normal exit status = %d",
+                    WEXITSTATUS(wstatus));
+        }
+
+        if (WIFSIGNALED(wstatus)) {
+            sprintf(buf + strlen(buf), " The process was killed by signal %d",
+                    WTERMSIG(wstatus));
+        }
+        strcat(buf, NL);
+        SendToAll(buf, strlen(buf), NULL);
     }
 
     // Clean up connections
@@ -639,35 +715,34 @@ void OnSigTerm(int)
     shutdownServer = true;
 }
 
+void OnSigHup(int)
+{
+    PRINTF("SigHup received\n");
+    openLogFile();
+}
+
 // Fork the daemon and exit the parent
 void forkAndGo()
 {
     pid_t p;
     int fh;
-    struct stat statBuf;
-
-    fstat(1, &statBuf);      // Find out what stdout is
 
     if ((p = fork()) < 0) {  // Fork failed
         perror("Could not fork daemon process");
         exit(errno);
-    } else if (p > 0) {      // I am the parent (foreground command)
+
+    } else if (p > 0) {      // I am the PARENT (foreground command)
         if (!quiet) {
             fprintf(stderr, "%s: spawning daemon process: %ld\n", procservName, (long) p);
-            if (logFileFD == -1) {
-                if (S_ISREG(statBuf.st_mode))
-                    fprintf(stderr, "The open file on stdout will be used as a log file.\n");
-                else
-                    fprintf(stderr, "No log file specified and stdout is not a file "
-                            "- no log will be kept.\n" );
+            if (-1 == logFileFD) {
+                fprintf(stderr, "Warning: No log file%s specified.\n",
+                        logPort ? "" : " and no port for log connections");
             }
         }
         exit(0);
-    } else {                 // I am the child (background daemon)
-        procservPid = getpid();
 
-        // Use stdout as log file, if no file is set and stdout is a file
-        if (logFileFD == -1 && S_ISREG(statBuf.st_mode)) logFileFD = dup(1);
+    } else {                 // I am the CHILD (background daemon)
+        procservPid = getpid();
 
         // Redirect stdin, stdout, stderr to /dev/null
         char buf[] = "/dev/null";
@@ -678,11 +753,7 @@ void forkAndGo()
         close(fh);
 
         // Make sure we are not attached to a terminal
-        fh = open("/dev/tty", O_RDWR);
-        if (fh >= 0) {                // It's a terminal
-            ioctl(fh, TIOCNOTTY);     // Detach from /dev/tty
-            close(fh);
-        }
+        setsid();
     }
 }
 
@@ -692,7 +763,8 @@ void forkAndGo()
 int checkCommandFile(const char * command)
 {
     struct stat s;
-    mode_t min_permissions=S_IRUSR|S_IXUSR|S_IXGRP|S_IRGRP|S_IROTH|S_IXOTH;
+    mode_t must_perm   =         S_IXUSR        |S_IXGRP        |S_IXOTH;
+    mode_t should_perm = S_IRUSR|S_IXUSR|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH;
 
     PRINTF("Checking command file validity\n");
 
@@ -711,14 +783,60 @@ int checkCommandFile(const char * command)
         fprintf(stderr, "%s: %s is not a regular file\n", procservName, command);
         return -1;
     }
-    if (min_permissions == (s.st_mode & min_permissions)) return 0; // This is great!
-    // else
-    fprintf(stderr, "%s: Warning - Please change permissions on %s to at least -r-xr-xr-x\n"
-             "procServ may not be able to continue without execute permission!\n",
-             procservName, command);
+
+    if (must_perm != (s.st_mode & must_perm)) {
+        fprintf(stderr, "%s: Error - Please change permissions on %s to at least ---x--x--x\n"
+                "procServ is not able to continue without execute permission\n",
+                procservName, command);
+        return -1;
+    }
+
+    if (should_perm != (s.st_mode & should_perm)) {
+        fprintf(stderr, "%s: Warning - Please change permissions on %s to at least -r-xr-xr-x\n"
+                "procServ may not be able to continue without read and execute permissions\n",
+                procservName, command);
+        return 0;
+    }
+
     return 0;
 }
 
+void openLogFile()
+{
+    if (-1 != logFileFD) {
+        close(logFileFD);
+    }
+    if (logFile) {
+        logFileFD = open(logFile, O_CREAT|O_WRONLY|O_APPEND, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
+        if (-1 == logFileFD) {         // Don't stop here - just go without
+            fprintf(stderr,
+                    "%s: unable to open log file %s\n",
+                    procservName, logFile);
+        } else {
+            PRINTF("Opened file %s for logging\n", logFile);
+        }
+    }
+}
+
+void ttySetCharNoEcho(bool set) {
+    static struct termios org_mode;
+    static struct termios mode;
+    static bool saved = false;
+
+    if (set && !saved) {
+        tcgetattr(0, &mode);
+        org_mode = mode;
+        saved = true;
+        mode.c_iflag &= ~IXON;
+        mode.c_lflag &= ~ICANON;
+        mode.c_lflag &= ~ECHO;
+        mode.c_cc[VMIN] = 1;
+        tcsetattr(0, TCSANOW, &mode);
+    } else if (saved) {
+        tcsetattr(0, TCSANOW, &org_mode);
+    }
+}
+
 connectionItem * connectionItem::head;
 // Globals:
 time_t procServStart; // Time when this IOC started
