@@ -1,6 +1,6 @@
 // Process server for soft ioc
 // David H. Thompson 8/29/2003
-// Ralph Lange 03/22/2010
+// Ralph Lange 02/28/2012
 // GNU Public License (GPLv3) applies - see www.gnu.org
 
 
@@ -37,6 +37,7 @@ extern char   killChar;
 extern char   toggleRestartChar;
 extern char   restartChar;
 extern char   quitChar;
+extern char   logoutChar;
 extern int    killSig;
 extern char   infoMessage1[];
 extern char   infoMessage2[];
@@ -48,6 +49,8 @@ extern time_t holdoffTime;
 
 #define NL "\r\n"
 
+#define CTL_SC(c) c > 0 && c < 32 ? "^" : "", c > 0 && c < 32 ? c + 64 : c
+
 class connectionItem;
 
 extern time_t procServStart; // Time when this IOC started
@@ -63,12 +66,11 @@ void SendToAll(const char * message,int count,const connectionItem * sender);
 void AddConnection(connectionItem *);
 void DeleteConnection(connectionItem *ci);
 
-
 // connectionItems are made in class factories so none of the
 // constructors are public:
 
 // processFactory creates the process that we are managing
-connectionItem * processFactory(int argc,char * argv[]);
+connectionItem * processFactory(char *exe, char *argv[]);
 bool processFactoryNeedsRestart(); // Call to test status of the server process
 void processFactorySendSignal(int signal);
 
@@ -105,6 +107,7 @@ public:
 
     // Return false unless you are the process item (processClass overloads)
     virtual bool isProcess() const { return false; }
+    virtual bool isLogger() const { return _readonly; }
 
 protected:
     connectionItem ( int fd = -1, bool readonly = false );
