@@ -1,6 +1,6 @@
 // Process server for soft ioc
 // David H. Thompson 8/29/2003
-// Ralph Lange 03/22/2010
+// Ralph Lange 04/13/2012
 // GNU Public License (GPLv3) applies - see www.gnu.org
 
 #include <unistd.h>
@@ -48,7 +48,7 @@ bool processFactoryNeedsRestart()
     return true;
 }
 
-connectionItem * processFactory(int argc, char *argv[])
+connectionItem * processFactory(char *exe, char *argv[])
 {
     char buf[512];
     time(&IOCStart); // Remember when we did this
@@ -63,7 +63,7 @@ connectionItem * processFactory(int argc, char *argv[])
             SendToAll( buf, strlen(buf), 0 );
         }
 
-        connectionItem *ci = new processClass( argc, argv );
+        connectionItem *ci = new processClass(exe, argv);
         PRINTF("Created new child connection (processClass %p)\n", ci);
 	return ci;
     }
@@ -111,8 +111,8 @@ processClass::~processClass()
 // This forks and
 //    parent: sets the minimum time for the next restart
 //    child:  sets the coresize, becomes a process group leader,
-//            and does an execv() with the command
-processClass::processClass(int argc,char * argv[])
+//            and does an execvp() with the command
+processClass::processClass(char *exe, char *argv[])
 {
     _runningItem=this;
     struct rlimit corelimit;
@@ -136,7 +136,8 @@ processClass::processClass(int argc,char * argv[])
         sprintf(infoMessage2, "@@@ Child \"%s\" PID: %ld" NL, childName, (long) _pid);
 
         sprintf(buf, "@@@ The PID of new child \"%s\" is: %ld" NL, childName, (long) _pid);
-        strcat(buf, "@@@ @@@ @@@ @@@ @@@" NL);
+        SendToAll( buf, strlen(buf), this );
+        strcpy(buf, "@@@ @@@ @@@ @@@ @@@" NL);
         SendToAll( buf, strlen(buf), this );
     }
     else                                    // I am the child
@@ -151,7 +152,7 @@ processClass::processClass(int argc,char * argv[])
             fprintf( stderr, "%s: child could not chdir to %s, %s\n",
                      procservName, chDir, strerror(errno) );
         } else {
-            execvp(*argv,argv);                         // execv()
+            execvp(exe, argv);              // execvp()
         }
 
 	// This shouldn't return, but did...
@@ -180,7 +181,7 @@ void processClass::readFromFd(void)
     }
 }
 
-// Send characters to clients
+// Sanitize buffer, then send characters to child
 int processClass::Send( const char * buf, int count )
 {
     int status = 0;
@@ -188,20 +189,6 @@ int processClass::Send( const char * buf, int count )
     char buf3[LINEBUF_LENGTH+1];
     char *buf2 = buf3;
 
-    // Scan input for commands
-    for ( i = 0; i < count; i++ ) {
-        if ( toggleRestartChar && buf[i] == toggleRestartChar ) {
-            autoRestart = ! autoRestart;
-            char msg[128];
-            sprintf ( msg, NL "@@@ Toggled auto restart to %s" NL,
-                      ( autoRestart ? "ON" : "OFF" ) );
-            SendToAll ( msg, strlen ( msg ), this );
-        }
-        if ( killChar && buf[i] == killChar ) {
-            PRINTF ("Got a kill command\n");
-            processFactorySendSignal( killSig );
-        }
-    }
                                 // Create working copy of buffer
     if ( count > LINEBUF_LENGTH ) buf2 = (char*) calloc (count + 1, 1);
     buf2[0] = '\0';
